import java.awt.Color;
import java.awt.Graphics;
import java.awt.Polygon;
import java.util.Random;
import java.util.Vector;

import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.WindowConstants;

import JTwilight.Transaction;
import JTwilight.GlobalVar;



public class DiningPhilosophers {
	static int NUM_PHILOSOPHERS = 19;
	static Vector<GlobalVar<Boolean>> table;
	static Philosopher[] philosophers;
	static AnimationPanel a;
	
	public static void main(String args[]) {
		initTable();
		initAnimation();
		initPhilosophers();
		startPhilosophers();
		startAnimation();
	}

	private static void startAnimation() {
		Animator anim = new Animator();
		anim.start();
	}

	private static void initAnimation() {
		JFrame f = new JFrame("Animation");
		f.setDefaultCloseOperation(WindowConstants.EXIT_ON_CLOSE);
		a = new AnimationPanel();
		f.add(a);
		f.setLocation(100, 100);
		f.setSize(400, 400);
		f.setVisible(true);
	}
	
	
	private static void startPhilosophers() {
		for(Philosopher philosopher : philosophers) {
			philosopher.start();
		}
	}

	private static void initPhilosophers() {
		philosophers = new Philosopher[NUM_PHILOSOPHERS];
		for (int i = 0; i < philosophers.length; i++) {
			philosophers[i] = new Philosopher(i);
		}
	}

	private static void initTable() {
		table = new Vector<GlobalVar<Boolean>>(NUM_PHILOSOPHERS);
		for (int i = 0; i < NUM_PHILOSOPHERS; i++) {
			table.add(new GlobalVar<Boolean> (Boolean.valueOf(true)));					
		}
	}
	
	private static class Philosopher extends Thread {
		int id;
		Random random = new Random();
		
		public Philosopher (int place) {
			id = place;
		}
		
		@Override
		public void run() {
			while (true) {
				think();
				eat();
			}
		}
	
		private void eat() {
			System.out.println("Philosopher "+ id +" is hungry.");
			getForks();
			try {
				Thread.sleep(random.nextInt(100) * 100);
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
			putForks();		
		}

		private void putForks() {
			Transaction.execute(new Transaction.TransactionBody<Void>() {
				public void transactional() {
					stm.write(table.get(id),Boolean.valueOf(true));
					stm.write(table.get((id +1) % NUM_PHILOSOPHERS),Boolean.valueOf(true));
				}
				public Void twilight() {
					System.out.println("Philosopher "+ id +" finished eating.");
					return null;
				}
			},null);
		}
		
		private void getForks() {
			Transaction.execute(new Transaction.TransactionBody<Void>() {
				public void transactional() {
					getFork(id);
					getFork((id + 1) % NUM_PHILOSOPHERS);
				}
				
				private void getFork(int id) {
					boolean available = stm.read(table.get(id));
					if (available) {
						stm.write(table.get(id),Boolean.valueOf(false));
					} else {
						stm.restart(stm);
					}
				}
				
				public Void twilight() {
					System.out.println("Philosopher "+ id +" is eating.");
					return null;
				}
				
			},null);
		}

		private void think() {
			System.out.println("Philosopher "+id+" is thinking.");
			try {
				Thread.sleep(random.nextInt(100) * 100);
			} catch (InterruptedException e) {}
		}
		
	}
	
	private static class AnimationPanel extends JPanel {
		private static final long serialVersionUID = 1L;
		Polygon p;
		int[] xs;
		int[] ys;
 		
		public AnimationPanel() {
			int n = table.size();
			xs = new int[n];
			ys = new int[n];
			
			for (int i = 0; i < n; i++){
				xs[i] = (int) Math.round(Math.cos(2*Math.PI/n * i) * 70.0 + 200.0); 
				ys[i] =	(int) Math.round(Math.sin(2*Math.PI/n * i) * 70.0 + 200.0);
			}
			p = new Polygon(xs,ys,n);
		}
		
		public synchronized void paintComponent (final Graphics g) {
			
			Transaction.execute(new Transaction.TransactionBody<Void>() {
				int n = table.size();
				boolean[] local_table = new boolean[n];
				public void transactional() {
					for (int i = 0; i < n; i++) {
						local_table[i] = stm.read(table.get(i));
					}
				}

				public Void twilight() {
					AnimationPanel.super.paintComponent(g);
					g.fillPolygon(p);
					g.setColor(Color.red);
					for (int i = 0; i < n; i ++) {
						if (!local_table[i]) {
							g.fillOval(xs[i]-4, ys[i]-4, 10, 10);
						}
					}
					return null;
				}
			}, null);
		}
	}	
	public static class Animator extends Thread {
		public void run() {
			while(true) {
				a.repaint();
				try {
					Thread.sleep(60);
				} catch (InterruptedException e) {
				}
			}
		}
		
	}
}
