import JTwilight.Transaction;
import JTwilight.GlobalVar;


public class RBTree {
	public enum Color {
		RED, BLACK
	};

	volatile GlobalVar<Node> root = new GlobalVar<Node>(null);

	private static class Node {
		volatile GlobalVar<Color> color;
		volatile GlobalVar<Integer> key;
		volatile GlobalVar<Node> parent;
		volatile GlobalVar<Node> left;
		volatile GlobalVar<Node> right;

		public Node(Node p) {
			color = new GlobalVar<Color>(Color.BLACK);
			key = null;
			parent = new GlobalVar<Node>(p);
			left = null;//new GlobalVar<Node>(null);
			right = null;//new GlobalVar<Node>(null);			
		}
		
		public Node(int i) {
			color = new GlobalVar<Color>(Color.RED);
			key = new GlobalVar<Integer>(i);
			parent = new GlobalVar<Node>(null);
			Node dummy = new Node(this);
			left = new GlobalVar<Node>(dummy);
			right = new GlobalVar<Node>(dummy);
		}

		public String toString(Transaction<?> stm) {
			String s = "";
			//dummy var
			if (key == null) {
				return s;
			}
			Node nleft = stm.read(left);
			Node nright = stm.read(right);
			if (nleft != null) {
				s += nleft.toString(stm) + " ";
			}
			String parent_key = "";
			if (stm.read(parent)  != null) {
				parent_key = Integer.toString(stm.read(stm.read(parent).key));
			}
			s += "(" + stm.read(key) + "," + stm.read(color) +"," + parent_key + ") ";
			if (nright != null) {
				s += nright.toString(stm) + " ";
			}
			return s;
		}
	}

	Node grandparent(Node n, Transaction<?> stm) {
		return stm.read(stm.read(n.parent).parent);
	}

	Node uncle(Node n, Transaction<?> stm) {
		Node granny = grandparent(n, stm);
		return (stm.read(n.parent) == stm.read(granny.left) ? stm.read(granny.right) : stm.read(granny.left));
	}

	private void reorder(Node n, Transaction<?> stm) {
		
		Node father = stm.read(n.parent);
		// 1. Case: inserted the root
		if (father == null) {
			stm.write(n.color, Color.BLACK);
			return;
		}
		// 2. Case: father is black
		if (stm.read(father.color) == Color.BLACK) {
			return;
		}
		// 3.Case: father and uncle are red
		Node granny = grandparent(n, stm);
		Node uncle = uncle(n, stm);
		if (uncle != null && stm.read(uncle.color) == Color.RED) {
			stm.write(father.color, Color.BLACK);
			stm.write(uncle.color, Color.BLACK);
			stm.write(granny.color, Color.RED);
			reorder(granny, stm);
			return;
		}
		//4. Case
		if (n == stm.read(father.right) && stm.read(n.parent) == stm.read(granny.left)) {
			rotate_left(n.parent, stm);
			n = stm.read(n.left);
			
		} else if (n == stm.read(father.left) && stm.read(n.parent) == stm.read(granny.right)) {
			rotate_right(n.parent, stm);
			n = stm.read(n.right);			
		}
		//5. case
		father = stm.read(n.parent);
		granny = grandparent(n,stm);
		stm.write(father.color, Color.BLACK);
		stm.write(granny.color, Color.RED);
		if (n == stm.read(father.left) && stm.read(n.parent) == stm.read(granny.left)) {
			rotate_right(father.parent, stm);
		} else {
			rotate_left(father.parent, stm);
		}
	}

	
	private void rotate_left(GlobalVar<Node> gn, Transaction<?> stm) {
		Node n = stm.read(gn);
		Node child = stm.read(n.right);
		stm.write(n.right,stm.read(child.left));
		if (stm.read(n.right) != null) {
			stm.write(stm.read(n.right).parent, n);
		}
		
		stm.write(child.left, n);
		
		if (stm.read(n.parent) == null) {
			stm.write(this.root, child);
			stm.write(child.parent,null);
		} else {
			if (stm.read(stm.read(n.parent).left) == n) {
				stm.write(stm.read(n.parent).left,child);
			} else {
				stm.write(stm.read(n.parent).right,child);
			}
			stm.write(child.parent, stm.read(n.parent));

		}
		stm.write(n.parent, child);
	}

	private void rotate_right(GlobalVar<Node> gn, Transaction<?> stm) {
		Node n = stm.read(gn);
		Node child = stm.read(n.left);
		stm.write(n.left,stm.read(child.right));
		if (stm.read(n.left) != null) {
			stm.write(stm.read(n.left).parent, n);
		}
		
		stm.write(child.right, n);
		
		if (stm.read(n.parent) == null) {
			stm.write(this.root, child);
			stm.write(child.parent,null);
		} else {
			if (stm.read(stm.read(n.parent).left) == n) {
				stm.write(stm.read(n.parent).left,child);
			} else {
				stm.write(stm.read(n.parent).right,child);
			}
		stm.write(child.parent, stm.read(n.parent));

		}
		stm.write(n.parent, child);
	}

	
	public Node search(int search_key, Transaction<?> stm) {
		Node current = stm.read(root);
		
		
		while (current != null && current.key != null) {
			if (stm.read(current.key) == search_key) {
				return current;
			} else {
				if (stm.read(current.key) > search_key) {
					current = stm.read(current.left);
				} else {
					current = stm.read(current.right);
				}
			}
		}
		return null;

	}
    
    public Node search(final int key) {
		return Transaction.execute(new Transaction.TransactionBody<Node>() {
            private Node node;
			public void transactional() {
				node = RBTree.this.search(key, stm);
			}

			public Node twilight() {
				return node;
			}
		}, null);
	}

	public void insert(int key, Transaction<?> stm) {
		Node n = new Node(key);
		
		Node current = stm.read(root);
		if (current == null) {
			// first node of the RBTree
			stm.write(root, n);
		} else {
			do {
				if (stm.read(current.key) == key) {
					// key already exists
					return;
				}
				if (stm.read(current.key) > key) {
					if (stm.read(current.left).key != null) {
						current = stm.read(current.left);
					} else {
						stm.write(n.parent, current);
						stm.write(current.left,n);
						break;
					}
				} else {
					if (stm.read(current.right).key != null) {
						current = stm.read(current.right);
					} else {
						stm.write(n.parent, current);
						stm.write(current.right,n);
						break;
					}
				}
			} while (true);
		}
		// re-balancing to re-establish the RBTree invariants
		reorder(n, stm);
	}

	public void insert(final int i) {
		Transaction.execute(new Transaction.TransactionBody<Void>() {
			public void transactional() {
				RBTree.this.insert(i, stm);
			}

			public Void twilight() {
				return null;
			}
		}, null);
	}
	
	//aka symmetric predecessor
	private Node nextSmallerKey (Node n,Transaction<?> stm) {
		Node current = stm.read(n.left);
		while (stm.read(current.right).key != null) {
			current = stm.read(current.right);
		}
		return current;
	}
	
	private boolean isLeaf(Node n, Transaction<?> stm) {
		return n.key == null;
	}
	
	private GlobalVar<Node> sibling (Node n, Transaction<?> stm) {
		if (n == stm.read(stm.read(n.parent).right)) {
			return stm.read(n.parent).left;
		} else {
			return stm.read(n.parent).right;
		}
	}
	
	public void delete(int key,Transaction<?> stm) {
		Node n = this.search(key, stm);
		if (n == null) {
			throw new IllegalArgumentException("Key to be deleted was not found");
		}
		if (stm.read(n.left).key != null && stm.read(n.right).key != null) {
			Node substitute = nextSmallerKey(n,stm);
			
			//copy substitute's values into the to be deleted node and delete the substitute
			int subst_key = stm.read(substitute.key);
			stm.write(n.key, subst_key);
			n = substitute;
		}
		
		Node child;
		if (isLeaf(stm.read(n.right),stm)) {
			child = stm.read(n.left);
		} else {
			child = stm.read(n.right);
		}
		//replace the node with its child
		Node parent = stm.read(n.parent);
		
		if (parent == null) {
			stm.write(this.root,child);
			stm.write(child.parent, null);
			
		} else {
			if (stm.read(parent.left) == n) {
				stm.write(parent.left, child);
			} else {
				stm.write(parent.right, child);
			}
			stm.write(child.parent, parent);
		}
		
		if (stm.read(n.color) == Color.BLACK) {
			if (stm.read(child.color) == Color.RED) {
				stm.write(child.color, Color.BLACK);
				return;
			} else {
				establishInvariants(child,stm);
			}
		}
	}
	

	private void establishInvariants(Node n, Transaction<?> stm) {
		//Case 1: root 
		Node parent = stm.read(n.parent);
		if ( parent == null) {
			return;
		}
		
		GlobalVar<Node> gsibling = sibling(n,stm);
		Node sibling = stm.read(gsibling); 
		//Case 2
		if (stm.read(sibling.color) == Color.RED) {
			stm.write(parent.color, Color.RED);
			stm.write(sibling.color, Color.BLACK);
			if (n == stm.read(parent.left)) {
				rotate_left(n.parent,stm);
			} else {
				rotate_right(n.parent,stm);
			}
		}
		
		parent = stm.read(n.parent);
		sibling = stm.read(sibling(n,stm));
		
		//Case 3
		if (stm.read(parent.color) == Color.BLACK && 
				stm.read(sibling.color) == Color.BLACK &&
				stm.read(stm.read(sibling.left).color) == Color.BLACK &&
				stm.read(stm.read(sibling.right).color) == Color.BLACK ) {
			stm.write(sibling.color, Color.RED);
			establishInvariants(parent,stm);
			return;
		}
		//Case 4
		if (stm.read(parent.color) == Color.RED && 
				stm.read(sibling.color) == Color.BLACK &&
				stm.read(stm.read(sibling.left).color) == Color.BLACK &&
				stm.read(stm.read(sibling.right).color) == Color.BLACK ) {
			stm.write(sibling.color, Color.RED);
			stm.write(parent.color, Color.BLACK);
			return;
		}
		//Case 5
		if (n == stm.read(parent.left) && 
				stm.read(sibling.color) == Color.BLACK &&
				stm.read(stm.read(sibling.left).color) == Color.RED &&
				stm.read(stm.read(sibling.right).color) == Color.BLACK ) {
			stm.write(sibling.color, Color.RED);
			stm.write(stm.read(sibling.left).color,Color.BLACK);
			rotate_right(sibling(n,stm),stm);
		} 
		else if (n == stm.read(parent.right) && 
				stm.read(sibling.color) == Color.BLACK &&
				stm.read(stm.read(sibling.right).color) == Color.RED &&
				stm.read(stm.read(sibling.left).color) == Color.BLACK ) {
			stm.write(sibling.color, Color.RED);
			stm.write(stm.read(sibling.right).color,Color.BLACK);
			rotate_left(sibling(n,stm),stm);
		}
		//Case 6
		sibling = stm.read(sibling(n,stm));
		parent = stm.read(n.parent);
		stm.write(sibling.color, stm.read(parent.color));
		stm.write(parent.color, Color.BLACK);
		if (n == stm.read(parent.left)) {
			stm.write(stm.read(sibling.right).color, Color.BLACK);
			rotate_left(n.parent,stm);
		} else {
			stm.write(stm.read(sibling.left).color, Color.BLACK);
			rotate_right(n.parent,stm);
		}		
	}

	public void delete(final int i) {
		Transaction.execute(new Transaction.TransactionBody<Void>() {
			public void transactional() {
                RBTree.this.delete(i, stm);
			}

			public Void twilight() {
				return null;
			}
		}, null);
	}

	
	public int depth () {
		final RBTree tree = this;
		return Transaction.execute(new Transaction.TransactionBody<Integer>() {
			int i;
			public void transactional() {
					i = tree.depth(stm);
			}

			public Integer twilight() {
				return i;
			}
		}, null);
	}
	
	protected int depth (Transaction<?> stm) {
		return depth(stm.read(root),stm);
	}

	private int depth(Node n,Transaction<?> stm) {
		if (n == null || n.key == null) {
			return 0;
		} else {
			return 1 + Math.max(depth(stm.read(n.left),stm), depth(stm.read(n.right),stm));
		}
	}

	public String toString(Transaction<?> stm) {
		Node nroot = stm.read(root);
		if (nroot == null) {
			return "";
		} else {
			return nroot.toString(stm);
		}

	}

	public String toString() {

		final RBTree tree = this;
		return Transaction.execute(new Transaction.TransactionBody<String>() {
			String s;

			public void transactional() {
				s = tree.toString(stm);
			}

			public String twilight() {
				return s;
			}
		}, null);
	}

}
