

public class RBTreeSequential {
	public enum Color {
		RED, BLACK
	};

	Node root = null;

	private static class Node {
		Color color;
		int key;
		Node parent;
		Node left;
		Node right;

		public Node(int i) {
			color = Color.RED;
			key = i; 
			parent = null;
			left = null;
			right = null;
		}

		public String toString() {
			String s = "";
			Node nleft = left;
			Node nright = right;
			if (nleft != null) {
				s += nleft.toString() + " ";
			}
			String parent_key = "";
			if (parent != null) {
				parent_key = Integer.toString(parent.key);
			}
			s += "(" + key + "," + color +"," + parent_key + ") ";
			if (nright != null) {
				s += nright.toString() + " ";
			}
			return s;
		}
	}

	Node grandparent(Node n) {
		return n.parent.parent;
	}

	Node uncle(Node n) {
		Node granny = grandparent(n);
		return (n.parent == granny.left ? granny.right : granny.left);
	}

	private void reorder(Node n) {
		
		Node father = n.parent;
		// 1. Case: inserted the root
		if (father == null) {
			n.color = Color.BLACK;
			return;
		}
		// 2. Case: father is black
		if (father.color == Color.BLACK) {
			return;
		}
		// 3.Case: father and uncle are red
		Node granny = grandparent(n);
		Node uncle = uncle(n);
		if (uncle != null && uncle.color == Color.RED) {
			father.color = Color.BLACK;
			uncle.color = Color.BLACK;
			granny.color = Color.RED;
			reorder(granny);
			return;
		}
		//4. Case
		if (n == father.right && n.parent == granny.left) {
			rotate_left(n.parent);
			n = n.left;
			
		} else if (n == father.left && n.parent == granny.right) {
			rotate_right(n.parent);
			n = n.right;			
		}
		//5. case
		father = n.parent;
		granny = grandparent(n);
		father.color = Color.BLACK;
		granny.color = Color.RED;
		if (n == father.left && n.parent == granny.left) {
			rotate_right(father.parent);
		} else {
			rotate_left(father.parent);
		}
	}

	private void rotate_right(Node n) {
		Node left_child = n.left;
		n.left = left_child.right;
		if (n.left != null) {
			n.left.parent = n;
		}		
		left_child.right = n;
		
		if (n.parent == null) {
			this.root = left_child;
			left_child.parent = null;
		} else {
			left_child.parent = n.parent;
			n.parent.left = left_child;
		}
		n.parent = left_child;
	}

	private void rotate_left(Node n) {
		Node right_child = n.right;
		n.right = right_child.left;
		if (n.right != null) {
			n.right.parent = n;
		}
		
		right_child.left = n;
		
		if (n.parent == null) {
			this.root = right_child;
			right_child.parent = null;
		} else {
			n.parent.right = right_child;
			right_child.parent = n.parent;

		}
		n.parent = right_child;
	}

	public Node search(int search_key) {
		Node current = root;
		while (current != null) {
			if (current.key == search_key) {
				return current;
			} else {
				if (current.key > search_key) {
					current = current.left;
				} else {
					current = current.right;
				}
			}
		}
		return null;

	}

	public void insert(int key) {
		Node n = new Node(key);
		
		Node current = root;
		if (current == null) {
			// first node of the RBTree
			root = n;
		} else {
			do {
				if (current.key == key) {
					// key already exists
					return;
				}
				if (current.key > key) {
					if (current.left != null) {
						current = current.left;
					} else {
						n.parent = current;
						current.left = n;
						break;
					}
				} else {
					if (current.right != null) {
						current = current.right;
					} else {
						n.parent = current;
						current.right = n;
						break;
					}
				}
			} while (true);
		}
		// re-balancing to re-establish the RBTree invariants
		reorder(n);
	}

	
	public int depth () {
		return depth(root);
	}

	private int depth(Node n) {
		if (n == null) {
			return 0;
		} else {
			return 1 + Math.max(depth(n.left), depth(n.right));
		}
	}

	public String toString() {
		Node nroot = root;
		if (nroot == null) {
			return "";
		} else {
			return nroot.toString();
		}

	}

}
