import java.util.Random;

import JTwilight.Transaction;
import JTwilight.GlobalVar;

public class SingleLinkedList {

	public static class Node {
		public volatile GlobalVar<Integer> value;
		public volatile GlobalVar<Node> next;

		public Node(int v) {
			value = new GlobalVar<Integer>(v);
			next = new GlobalVar<Node>(null);
		}
	}

	public final GlobalVar<Node> head = new GlobalVar<Node>(null);

	public boolean insert(int key, Transaction<?> stm) {
		Node newNode = new Node(key);
		Node current = stm.read(head);

		// if list is still empty, insert a new head
		if (current == null) {
			stm.write(head, newNode);
			return true;
		}

		// insert at the head of the list
		if (stm.read(current.value) > key) {
			stm.write(newNode.next, current);
			stm.write(head, newNode);
			return true;
		}

		// insert in the middle of the list
		while (current != null) {
			Node cnext = stm.read(current.next);
			if (cnext != null) {
				// key already there
				if (stm.read(cnext.value) == key) {
					return false;
				}
				if (stm.read(cnext.value) < key) {
					current = cnext;
				} else {
					stm.write(newNode.next, cnext);
					stm.write(current.next, newNode);
					return true;
				}
			} else {
				// insert at the end of the list
				stm.write(current.next, newNode);
				return true;
			}
		}
		// this code is unreachable
		assert (false);
		return false;
	}

	public String toString(Transaction<?> stm) {
		String s = "Values in List: ";

		GlobalVar<Node> globaln = head;
		while (globaln != null) {
			Node n = stm.read(globaln);
			if (n != null) {
				s += " " + stm.read(n.value);
				globaln = n.next;
			} else {
				globaln = null;
			}
		}
		return s;
	}


	// Testing

	static int INSERTED_ELEMENTS = 5;
	static int UPPER_BOUND = 100000;
	static int NUM_THREADS = 8;
	static Random randomGen = new Random();

	public static class Tester extends Thread {
		final GlobalVar<SingleLinkedList> l;

		public Tester(GlobalVar<SingleLinkedList> l) {
			this.l = l;
		}

		public void run() {
			for (int i = 0; i < INSERTED_ELEMENTS; i++) {
				final int new_value = randomGen.nextInt(UPPER_BOUND) + 1;
				Transaction.execute(new Transaction.TransactionBody<Void>() {

					public void transactional() {
						SingleLinkedList list = stm.read(l);
						list.insert(new_value, stm);
					}

					public Void twilight() {
						return null;
					}
				}, null);
			}
		}
	}

	public static void main(String[] args) {
		final GlobalVar<SingleLinkedList> globallist = new GlobalVar<SingleLinkedList>(
				new SingleLinkedList());

		Tester threads[] = new Tester[NUM_THREADS];

		System.out.println("Initializing transactions");
		for (int i = 0; i < NUM_THREADS; ++i) {
			threads[i] = new Tester(globallist);
		}
		System.out.println("Starting transactions");
		for (int i = 0; i < NUM_THREADS; ++i) {
			threads[i].start();
		}

		try {
			for (int i = 0; i < NUM_THREADS; ++i) {
				threads[i].join();
			}
		} catch (InterruptedException e) {
			System.out.println("Urgh");
		}
		
		Transaction.execute(new Transaction.TransactionBody<Void>() {
			String s;
			public void transactional() {
				SingleLinkedList list = stm.read(globallist);
				s = list.toString(stm);
			}

			public Void twilight() {
				System.out.println(s);
				return null;
			}
		}, null);
		
	}
}
