/* =============================================================================
 *
 * tl2.h
 *
 * Transactional Locking 2 software transactional memory
 *
 * =============================================================================
 *
 * Copyright (C) Sun Microsystems Inc., 2006.  All Rights Reserved.
 * Authors: Dave Dice, Nir Shavit, Ori Shalev.
 *
 * TL2: Transactional Locking for Disjoint Access Parallelism
 *
 * Transactional Locking II,
 * Dave Dice, Ori Shalev, Nir Shavit
 * DISC 2006, Sept 2006, Stockholm, Sweden.
 *
 * =============================================================================
 *
 * Modified by Chi Cao Minh (caominh@stanford.edu)
 *
 * See VERSIONS for revision history
 *
 * =============================================================================
 */

#ifndef TL2_H
#define TL2_H 1














#include <stdint.h>
#include "tmalloc.h"


#  include <setjmp.h>






typedef struct _Thread Thread;

#ifdef __cplusplus
extern "C" {
#endif





#  include <setjmp.h>
#  define SIGSETJMP(env, savesigs)      sigsetjmp(env, savesigs)
#  define SIGLONGJMP(env, val)          siglongjmp(env, val); assert(0)


typedef unsigned char  boolean;
typedef uintptr_t      vwLock;  /* (Version,LOCKBIT) */


/*
 * Prototypes
 */




void     TxStart       (Thread*, sigjmp_buf*, int*);

Thread*  TxNewThread   ();





void     TxFreeThread  (Thread*);
void     TxInitThread  (Thread*, long id);
int      TxBeginCommit (Thread*);
void     TxEndCommit   (Thread*);
void     TxAbort       (Thread*);
intptr_t TxLoad        (Thread*, volatile intptr_t*, boolean);
void     TxStore       (Thread*, volatile intptr_t*, intptr_t, boolean);
void     TxStoreLocal  (Thread*, volatile intptr_t*, intptr_t);
void     TxStoreUpdate (Thread* Self, volatile intptr_t* addr, intptr_t valu);
void     TxOnce        ();
void     TxShutdown    ();

void*    TxAlloc       (Thread*, size_t);
void     TxFree        (Thread*, void*);

#define WAS_INCOHERENT_READ(result) (result & 2)
#define WAS_UNABLE_TO_OBTAIN_LOCKS(result) (result & 4)
#define COMMIT_FAILED(result) (result != 0)

typedef struct _Incoherence {
    volatile intptr_t *Addr;
    boolean isIncoherent;
    intptr_t reserved;
    int zone;
} Incoherence;
Incoherence* NextIncoherence (Thread*, const Incoherence*, Incoherence*);

void MarkIncoherent(Thread*, volatile intptr_t*);
intptr_t GetReadsetValue(Thread*, volatile intptr_t*);
int TryLoadReadset(Thread*);
int LoadReadsetWithLocks(Thread*);
void RemoveFromReadset(Thread*, volatile intptr_t*);

void ForgetZone(Thread*, int);
void EnterZone(Thread*, int);
void LeaveZone(Thread*);

typedef struct _HistoryEntry {
    volatile vwLock   version;
    volatile intptr_t value;
} HistoryEntry;

typedef struct _History {
    volatile vwLock lock;
    volatile short max;
    volatile short use;
    volatile intptr_t head;
    volatile HistoryEntry *entries;
} History;


#ifdef __cplusplus
}
#endif


#endif /* TL2_H */


/* =============================================================================
 *
 * End of tl2.h
 *
 * =============================================================================
 */
